/*  app.cpp
 *
 *  The laserbreath application connects the ADC stream to the DAC stream
 *  and inserts a linear phase bandsplitter into the DAC stream.
 *  The sliders in the top half of the application window control
 *  the attenuation of each audio band and the bottom row of sliders
 *  controls a delay applied to each band.
 *
 *  To test the application put an audio CD in the CD drive and use
 *  the CD Player application to play the CD.  Using the sound panel,
 *  select the CD as the input source, mute the loopback and CD output
 *  mix controls so that you are listening to the DAC as the ADC is
 *  reading from the CD.
 */

#include <OS.h>
#include <Alert.h>
#include <stdio.h>

#include "app.h"
#include "nino.h"

#define NBANDS 13					// Number of audio bands
#define MAX_DELAY 0.5				// Maximum delay (in seconds)

#define HALF_SPEED TRUE				// Do DSP at half sampling rate

#define SLIDER_WIDTH 14				// Width of a slider stripe (in pixels)
#define SLIDER_X(I) (35 * (I) + 50)	// horizontal position of slider i

int
main ()
{
  LBApp* app = new LBApp(NBANDS, MAX_DELAY);	// start bandsplitter
  Nino* nino = new Nino();						// then connect ADC to DAC

  app->Run();

  delete nino;
  delete app;
  return 0;
}

static bool
buffer_handler (void* arg, char* buf, long count)
{
  ((Splitter*) arg)->ProcessBuffer ((frame*) buf, count);
  return TRUE;
}

LBApp::LBApp(int n_bands, float max_delay_secs)
  : BApplication('LSRB')
{
  // set streams to 16-bit stereo and run bandsplitter

  subscriber = new BAudioSubscriber ("Splitter");
  subscriber->Subscribe (B_DAC_STREAM, B_SHARED_SUBSCRIBER_ID, TRUE);
  subscriber->SetDACSampleInfo (2, 2, B_BIG_ENDIAN, B_LINEAR_SAMPLES);
  subscriber->SetADCSampleInfo (2, 2, B_BIG_ENDIAN, B_LINEAR_SAMPLES);
  if (HALF_SPEED)
	processingRate = subscriber->SamplingRate() / 2;
  else
	processingRate = subscriber->SamplingRate();
  splitter = new Splitter (n_bands, max_delay_secs*processingRate, HALF_SPEED);
  subscriber->EnterStream (NULL, FALSE, splitter, buffer_handler, NULL, TRUE);

  // create window and main view

  BRect rect(200, 100, 180 + SLIDER_X(NBANDS) + SLIDER_WIDTH, 550);
  LBWindow* win = new LBWindow (rect, "Laserbreath");
  win->Lock();

  rect.OffsetTo (B_ORIGIN);
  LBView* view = new LBView (rect, "a view");
  win->AddChild (view);

  // add sliders

  rect.Set (0, 0, SLIDER_WIDTH, 200);
  for (int i = 0; i < NBANDS; i++) {
	rect.OffsetTo (SLIDER_X(i), 10);
	LBSlider* slider = new GainSlider (rect, "", i);
	view->AddChild (slider);
	slider->SetValue (0.7);

	rect.OffsetTo (SLIDER_X(i), 240);
	slider = new DelaySlider (rect, "", i);
	view->AddChild (slider);
	slider->SetValue ((float) i / NBANDS / 2);
  }

  win->Unlock();
  win->Show();
}

LBApp::~LBApp()
{
  subscriber->ExitStream (TRUE);
  subscriber->Unsubscribe();
}

void
LBApp::AboutRequested()
{
  BAlert* alert = new BAlert("", "Bandsplitter demo app version 0.1", "OK");
  if (alert)
	alert->Go();
}

void
LBView::Draw (BRect r)
{
  SetFontName ("Emily");
  SetFontSize (9);
  SetLowColor (ViewColor ());
  for (int i = 0; i < NBANDS; i++) {
	float fs = ((LBApp*) be_app)->processingRate;
	float frequency = fs * i / (2.0 * (NBANDS - 1)) / 1000;
	char buf[12];
	sprintf (buf, "%2.1f", frequency);
	DrawString (buf, BPoint (SLIDER_X(i) + 1, 229));
  }
  DrawString ("Level", BPoint (10, 114));
  DrawString ("kHz", BPoint (13, 229));
  DrawString ("Delay", BPoint (10, 344));
}

void
GainSlider::ValueChanged (float val)
{
  LBApp* app = (LBApp*) be_app;
  app->splitter->SetGain (band, pow (10, 2 * (val - 1)));
}

void
DelaySlider::ValueChanged (float val)
{
  LBApp* app = (LBApp*) be_app;
  app->splitter->SetDelay (band, (int)(MAX_DELAY * app->processingRate * val));
}
