/*  Laserbreath -
 *    Rabiner / Gold development of recursive FFT Goertzel algorithm;
 *    linear phase bandsplitting combined with delays.
 *    Implements scheme described by Andy Moorer in JAES v34 #3.
 */

#include "splitter.h"
#include <stdio.h>
#include <math.h>

/* Splitter (int n_bands, int max_delay, bool halfspeed, double radius)

	n_bands is the number of audio bands which will divide
   		the spectrum between 0Hz and half the sampling rate.

	max_delay is size of the input delay line in samples.

	halfspeed directs the bandsplitter to ignore every other input
		sample and reconstruct the output by linear interpolation.

	radius is a damping factor for the resonators.
*/

Splitter::Splitter (int n_bands, int max_delay, bool halfspeed, double radius)
{
  nBands = n_bands;
  nPoles = 2 * (nBands - 1);
  downsampleFactor = (halfspeed ? 2 : 1);
  maxDelay = max_delay / downsampleFactor;

  // initialize bands

  bands = new Band[nBands];
  for (int m = 0;  m < nBands; m++) {
	Band* b = &bands[m];
	SetDelay (m, 0);
	SetGain (m, 1.0);
	b->a1 = radius * cos (2 * M_PI * m / nPoles);
	b->a2 = radius * sin (2 * M_PI * m / nPoles);
	b->yL = 0;
	b->wL = 0;
	b->yR = 0;
	b->wR = 0;
	b->comb = new frame[nPoles];
	for (int i = 0; i < nPoles; i++)
	  b->comb[i] = zero_frame;
  }

  // initialize delay line

  delayBuf = new frame[maxDelay + 1];
  for (int i = 0; i <= maxDelay; i++)
	delayBuf[i] = zero_frame;

  delayPointer = 0;
  combPointer = 0;
  prevL = 0;
  prevR = 0;
}

Splitter::~Splitter ()
{
  for (int m = 0;  m < nBands; m++)
	delete [] bands[m].comb;
  delete [] bands;
  delete [] delayBuf;
}

void
Splitter::SetGain (int m, float gain)
{
  if (m > 0 && m < nBands - 1)
	gain *= 2;
  bands[m].gain = gain;
}

void
Splitter::SetDelay (int m, int delay)
{
  if (delay > maxDelay)
	delay = maxDelay;
  else if (delay < 0)
	delay = 0;
  bands[m].delay = delay;
}

/*	ProcessBuffer (frame* buf, int bufbytes)

	for each band if the input samples are X[n]
	then the comb filter output is C[n] = X[n] - X[n - nPoles]
	and the resonator output is
	  Y[n] = C[n] + a1 * Y[n - 1] + a2 * W[n - 1]
	where
	  W[n] = a2 * Y[n - 1] + a1 * W[n - 1]
*/

void
Splitter::ProcessBuffer (frame* buf, int bufbytes)
{
  dsp_reg divide_by_n_poles = (double) 1.0 / nPoles;
  int n_frames = bufbytes / sizeof (*buf);

  for (int i = 0; i < n_frames; i += downsampleFactor) {
	dsp_reg mixL = 0;						// for each input frame
	dsp_reg mixR = 0;
	delayBuf[delayPointer] = *buf;			// run delay line

	register Band* b = bands;
	for (int m = 0; m < nBands; m++, b++) {		// for each band
	  int index = delayPointer - b->delay;
	  if (index < 0)
		index += maxDelay + 1;

	  int comb_inL = delayBuf[index].L;			// run comb filter
	  int comb_inR = delayBuf[index].R;
	  int comb_outL = comb_inL - b->comb[combPointer].L;
	  int comb_outR = comb_inR - b->comb[combPointer].R;
	  b->comb[combPointer].L = comb_inL;
	  b->comb[combPointer].R = comb_inR;

	  if (m == 0) {								// run resonator
		b->yL = comb_outL + b->yL;				// first band
		b->yR = comb_outR + b->yR;
	  }
	  else if (m == nBands-1) {
		b->yL = comb_outL - b->yL;				// last band
		b->yR = comb_outR - b->yR;
	  }
	  else {
		dsp_reg prev_yL = b->yL;				// other bands
		dsp_reg prev_yR = b->yR;
		b->yL = comb_outL + b->a1 * prev_yL - b->a2 * b->wL;
		b->yR = comb_outR + b->a1 * prev_yR - b->a2 * b->wR;
		b->wL = b->a2 * prev_yL + b->a1 * b->wL;
		b->wR = b->a2 * prev_yR + b->a1 * b->wR;
	  }
	  mixL += b->gain * b->yL;					// mix output
	  mixR += b->gain * b->yR;
	}

	// scale and round

	register outL = (int) (mixL * divide_by_n_poles + 32768.5) - 32768;
	register outR = (int) (mixR * divide_by_n_poles + 32768.5) - 32768;

	if (downsampleFactor == 2) {

	  register interpL = (outL + prevL) >> 1;	// interpolate if necessary
	  register interpR = (outR + prevR) >> 1;
	  prevL = outL;
	  prevR = outR;

	  if (interpL > 32767)						// clip
		interpL = 32767;
	  else if (interpL < -32768)
		interpL = -32768;
	  buf->L = interpL;

	  if (interpR > 32767)
		interpR = 32767;
	  else if (interpR < -32768)
		interpR = -32768;
	  buf->R = interpR;

	  ++buf;
	}

	if (outL > 32767)							// clip
	  outL = 32767;
	else if (outL < -32768)
	  outL = -32768;
	buf->L = outL;

	if (outR > 32767)
	  outR = 32767;
	else if (outR < -32768)
	  outR = -32768;
	buf->R = outR;

	++buf;

	// update comb filter and delay line pointers

	if (++combPointer >= nPoles)
	  combPointer = 0;
	if (++delayPointer > maxDelay)
	  delayPointer = 0;
  }
}


